import os
import pycurl
import pytest
import shutil

from munch import Munch
from tempfile import mkdtemp, NamedTemporaryFile
from mock import patch, Mock, call

from . import base

from rpkglib.errors import RpkgError
from rpkglib import LookasideCache

class TestLookasideCache(base.TestCase):

    def setUp(self):
        super(TestLookasideCache, self).setUp()

    def tearDown(self):
        super(TestLookasideCache, self).tearDown()

    def test_get_download_url(self):
        lookaside = LookasideCache(
            path=self.cwd,
            download_url='http://localhost/repo/pkgs/%(repo_path)s/%(filename)s/%(hashtype)s/%(hash)s/%(filename)s',
            upload_url=None)

        result = lookaside.get_download_url('foo/bar', 'baz', 'sha256', 'hash')
        self.assertEqual(result, 'http://localhost/repo/pkgs/foo/bar/baz/sha256/hash/baz')

        result = lookaside.get_download_url('foo/bar/', 'baz baz', 'sha256', 'hash')
        self.assertEqual(result, 'http://localhost/repo/pkgs/foo/bar/baz%20baz/sha256/hash/baz%20baz')

        lookaside = LookasideCache(
            path=self.cwd,
            download_url='http://localhost/repo/pkgs/%(repo_path)s/%(filename)s/%(hashtype)s/%(hash)s/%(filename)s/%(kwarg)s',
            upload_url=None)

        result = lookaside.get_download_url('foo/bar', 'baz', 'sha256', 'hash', kwarg='kwarg')
        self.assertEqual(result, 'http://localhost/repo/pkgs/foo/bar/baz/sha256/hash/baz/kwarg')

    @patch('rpkglib.lookaside_cache.log')
    def test_add_download_certs(self, log):
        lookaside = LookasideCache(
            path=self.cwd,
            download_url=None,
            upload_url=None)

        c = Mock()
        lookaside.add_download_certs(c)
        c.setopt.assert_not_called()
        log.warning.assert_not_called()

        lookaside = LookasideCache(
            path=self.cwd,
            download_url=None,
            upload_url=None,
            ca_cert='ca.pem')

        c.reset_mock()
        lookaside.add_download_certs(c)
        c.setopt.assert_not_called()
        log.warning.assert_called()

        open('ca.pem', 'w').close()
        c.reset_mock()
        lookaside.add_download_certs(c)
        c.setopt.assert_called_with(pycurl.CAINFO, 'ca.pem')
        os.unlink('ca.pem')

    @patch('sys.stdout')
    @patch('rpkglib.lookaside_cache.log')
    @patch('rpkglib.lookaside_cache.pycurl.Curl')
    def test_download(self, curl_cls, log, stdout):
        tmpdir = mkdtemp()
        target = os.path.join(tmpdir, 'filename')

        lookaside = LookasideCache(
            path=self.cwd,
            download_url='https://pagure.io/rpkg-util',
            upload_url=None)

        curl = Mock()
        curl_cls.return_value = curl

        def getinfo(param):
            if param == pycurl.INFO_FILETIME:
                return 1551628560
            if param == pycurl.RESPONSE_CODE:
                return 200

        def perform():
            with open(target, 'w') as f:
                f.write('foo\n')

        curl.getinfo = getinfo
        curl.perform = perform
        lookaside.download(target, '', 'filename', 'sha256', 'b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f4850b878ae4944c')
        statinfo = os.stat(target)
        self.assertEqual(statinfo.st_mtime, 1551628560)

        curl.reset_mock()
        lookaside.download(target, '', 'filename', 'sha256', 'b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f4850b878ae4944c')
        log.debug.assert_called_with('File %s already downloaded.', target)
        os.unlink(target)

        def getinfo_invalid_response(param):
            if param == pycurl.INFO_FILETIME:
                return 1551628560
            if param == pycurl.RESPONSE_CODE:
                return 300

        curl.reset_mock()
        curl.getinfo = getinfo_invalid_response
        with pytest.raises(RpkgError):
            lookaside.download(target, '', 'filename', 'sha256', 'b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f4850b878ae4944c')

        def perform_invalid_content():
            with open(target, 'w') as f:
                f.write('bar\n')

        curl.getinfo = getinfo
        curl.perform = perform_invalid_content

        with pytest.raises(RpkgError):
            lookaside.download(target, '', 'filename', 'sha256', 'b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f4850b878ae4944c')

        shutil.rmtree(tmpdir)

    def test_hashsum(self):
        lookaside = LookasideCache(
            path=self.cwd,
            download_url=None,
            upload_url=None)

        file = NamedTemporaryFile(mode='w', delete=False)
        file.write('foo\n')
        file.close()

        result = lookaside.hashsum(file.name, 'sha512')
        self.assertEqual(result, '0cf9180a764aba863a67b6d72f0918bc131c6772642cb2dce5a34f0a702f9470ddc2bf125c12198b1995c233c34b4afd346c54a2334c350a948a51b6e8b4e6b6')

        result = lookaside.hashsum(file.name, 'sha256')
        self.assertEqual(result, 'b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f4850b878ae4944c')

        result = lookaside.hashsum(file.name, 'md5')
        self.assertEqual(result, 'd3b07384d113edec49eaa6238ad5ff00')

        os.unlink(file.name)

    def test_file_is_valid(self):
        lookaside = LookasideCache(
            path=self.cwd,
            download_url=None,
            upload_url=None)

        file = NamedTemporaryFile(mode='w', delete=False)
        file.write('foo\n')
        file.close()

        result = lookaside.file_is_valid(file.name, 'sha256', 'b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f4850b878ae4944c')
        self.assertTrue(result)

        result = lookaside.file_is_valid(file.name, 'md5', 'd3b07384d113edec49eaa6238ad5ff00')
        self.assertTrue(result)

        result = lookaside.file_is_valid(file.name, 'sha256', 'b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f4850b878ae4944d')
        self.assertFalse(result)

        result = lookaside.file_is_valid(file.name, 'md5', 'd3b07384d113edec49eaa6238ad5ff01')
        self.assertFalse(result)

        os.unlink(file.name)

    def test_read_source_entries(self):
        source_path = os.path.join(self.cwd, 'sources')

        file = open(source_path, 'w')
        file.write('SHA512 (prunerepo-1.15.tar.gz) = 623482')
        file.close()
        lookaside = LookasideCache(self.cwd, None, None)
        result = lookaside.read_source_entries()
        expected_result = [Munch({'filename': 'prunerepo-1.15.tar.gz',
                                 'hashtype': 'sha512',
                                 'hash': '623482'})]
        self.assertEqual(result, expected_result)

        file = open(source_path, 'w')
        file.write('SHA512 (prunerepo-1.15.tar.gz) = 623482\n')
        file.write('SHA512 (dist-git-1.11.tar.gz) = 7aacb6\n')
        file.close()
        lookaside = LookasideCache(self.cwd, None, None)
        result = lookaside.read_source_entries()
        expected_result = [
            Munch({'filename': 'prunerepo-1.15.tar.gz',
                   'hashtype': 'sha512',
                   'hash': '623482'}),
            Munch({'filename': 'dist-git-1.11.tar.gz',
                   'hashtype': 'sha512',
                   'hash': '7aacb6'})]
        self.assertEqual(result, expected_result)

        file = open(source_path, 'w')
        file.write('SHA512 invalid = err\n')
        file.close()
        lookaside = LookasideCache(self.cwd, None, None)
        with pytest.raises(RpkgError):
             lookaside.read_source_entries()

        os.unlink(source_path)

    def test_write_source_entries(self):
        lookaside = LookasideCache(
            path=self.cwd,
            download_url=None,
            upload_url=None)

        source_entries = [
            Munch({'filename': 'foo.tar.gz', 'hashtype': 'sha512', 'hash': '7aacb6'}),
            Munch({'filename': 'bar.tar.gz', 'hashtype': 'sha512', 'hash': '0cf918'}),
        ]

        self.assertNotIn('sources', os.listdir('.'))
        lookaside.write_source_entries(source_entries)
        self.assertIn('sources', os.listdir('.'))
        result = lookaside.read_source_entries()
        self.assertEqual(result, source_entries)

        lines = []
        with open(os.path.join(self.cwd, 'sources')) as sources_file:
            lines = [l.strip() for l in sources_file.readlines()]
        self.assertEqual(lines, ['SHA512 (foo.tar.gz) = 7aacb6', 'SHA512 (bar.tar.gz) = 0cf918'])

        source_entries = [
            Munch({'filename': 'foo.tar.gz', 'hashtype': 'sha512', 'hash': '7aacb6'}),
            Munch({'filename': 'bar.tar.gz', 'hashtype': 'sha512', 'hash': '0cf919'}),
        ]

        lookaside.write_source_entries(source_entries)
        result = lookaside.read_source_entries()
        self.assertEqual(result, [
            Munch({'filename': 'foo.tar.gz', 'hashtype': 'sha512', 'hash': '7aacb6'}),
            Munch({'filename': 'bar.tar.gz', 'hashtype': 'sha512', 'hash': '0cf918'}),
            Munch({'filename': 'bar.tar.gz', 'hashtype': 'sha512', 'hash': '0cf919'}),
        ])

        lookaside.write_source_entries(source_entries, append=False)
        result = lookaside.read_source_entries()
        self.assertEqual(result, source_entries)

    def test_make_source_entry(self):
        lookaside = LookasideCache(
            path=self.cwd,
            download_url=None,
            upload_url=None)

        file = NamedTemporaryFile(mode='w', delete=False)
        file.write('foo\n')
        file.close()

        result = lookaside.make_source_entry(file.name)
        self.assertEqual(result, Munch({
            'filename': os.path.basename(file.name),
             'hashtype': 'sha512',
             'hash': '0cf9180a764aba863a67b6d72f0918bc131c6772642cb2dce5a34f0a702f9470ddc2bf125c12198b1995c233c34b4afd346c54a2334c350a948a51b6e8b4e6b6'}))

        os.unlink(file.name)

    @patch('rpkglib.lookaside_cache.log')
    def test_add_upload_certs(self, log):
        lookaside = LookasideCache(
            path=self.cwd,
            download_url=None,
            upload_url=None)

        c = Mock()
        lookaside.add_upload_certs(c)
        c.setopt.assert_not_called()
        log.warning.assert_not_called()

        lookaside = LookasideCache(
            path=self.cwd,
            download_url=None,
            upload_url=None,
            client_key='client.key',
            client_cert='client.pem',
            ca_cert='ca.pem')

        c.reset_mock()
        lookaside.add_upload_certs(c)
        c.setopt.assert_not_called()
        self.assertEqual(log.warning.call_count, 3)

        open('client.key', 'w').close()
        open('client.pem', 'w').close()
        open('ca.pem', 'w').close()
        c.reset_mock()
        lookaside.add_upload_certs(c)
        c.setopt.assert_has_calls([
            call(pycurl.SSLKEY, 'client.key'),
            call(pycurl.SSLCERT, 'client.pem'),
            call(pycurl.CAINFO, 'ca.pem')])
        os.unlink('client.key')
        os.unlink('client.pem')
        os.unlink('ca.pem')

    @patch('sys.stdout')
    @patch('rpkglib.lookaside_cache.log')
    @patch('rpkglib.lookaside_cache.pycurl.Curl')
    @patch('rpkglib.lookaside_cache.BytesIO')
    def test_source_uploaded(self, bytes_cls, curl_cls, log, stdout):
        lookaside = LookasideCache(
            path=self.cwd,
            download_url=None,
            upload_url='https://pagure.io/repo/pkgs/upload.cgi')

        buf = Mock()
        bytes_cls.return_value = buf

        curl = Mock()
        curl_cls.return_value = curl

        def getinfo(param):
            if param == pycurl.RESPONSE_CODE:
                return 200

        def perform():
            buf.getvalue.return_value = b'Available'

        curl.perform = perform
        curl.getinfo = getinfo
        result = lookaside.source_uploaded('/repo_path', 'filename', 'sha512', 'b5bbd8')

        self.assertTrue(result)
        curl.setopt.assert_has_calls([
            call(pycurl.URL, 'https://pagure.io/repo/pkgs/upload.cgi'),
            call(pycurl.HTTPPOST, [
                ('filename', 'filename'),
                ('name', '/repo_path'),
                ('sha512sum', 'b5bbd8')]),
        ], any_order=True)

        def perform():
            buf.getvalue.return_value = b'Missing'

        curl.perform = perform
        result = lookaside.source_uploaded('/repo_path', 'filename', 'sha512', 'b5bbd8')

        self.assertFalse(result)
        def getinfo(param):
            if param == pycurl.RESPONSE_CODE:
                return 403

        curl.getinfo = getinfo
        with pytest.raises(RpkgError):
            lookaside.source_uploaded('/repo_path', 'filename', 'sha512', 'b5bbd8')

    @patch('sys.stdout')
    @patch('rpkglib.lookaside_cache.log')
    @patch('rpkglib.lookaside_cache.pycurl.Curl')
    @patch('rpkglib.lookaside_cache.BytesIO')
    @patch('rpkglib.lookaside_cache.os.path.getmtime')
    def test_upload_source(self, getmtime, bytes_cls, curl_cls, log, stdout):
        lookaside = LookasideCache(
            path=self.cwd,
            download_url=None,
            upload_url=None)

        with pytest.raises(RpkgError):
            lookaside.upload_source('/filename', '/repo_path', 'filename', 'sha512', 'b5bbd8')

        lookaside = LookasideCache(
            path=self.cwd,
            download_url=None,
            upload_url='https://pagure.io/repo/pkgs/upload.cgi')

        buf = Mock()
        bytes_cls.return_value = buf

        curl = Mock()
        curl_cls.return_value = curl

        lookaside.source_uploaded = Mock()
        lookaside.source_uploaded.return_value = True
        lookaside.upload_source('/filename', '/repo_path', 'filename', 'sha512', 'b5bbd8')
        curl.perform.assert_not_called()

        lookaside.source_uploaded.return_value = False

        def getinfo(param):
            if param == pycurl.RESPONSE_CODE:
                return 200

        curl.getinfo = getinfo
        getmtime.return_value = 1234

        lookaside.upload_source('/filename', '/repo_path', 'filename', 'sha512', 'b5bbd8')

        curl.setopt.assert_has_calls([
            call(pycurl.URL, 'https://pagure.io/repo/pkgs/upload.cgi'),
            call(pycurl.HTTPPOST, [
                ('file', (pycurl.FORM_FILE, '/filename')),
                ('mtime', '1234'),
                ('name', '/repo_path'),
                ('sha512sum', 'b5bbd8')]),
        ], any_order=True)
        curl.perform.assert_called()

        def getinfo(param):
            if param == pycurl.RESPONSE_CODE:
                return 403

        curl.getinfo = getinfo
        with pytest.raises(RpkgError):
            lookaside.upload_source('/filename', '/repo_path', 'filename', 'sha512', 'b5bbd8')

    def test_sources(self):
        source_path = os.path.join(self.cwd, 'sources')
        lookaside = LookasideCache(self.cwd, None, None)
        lookaside.read_source_entries = Mock()
        lookaside.download = Mock()

        with pytest.raises(RpkgError):
            lookaside.sources(self.cwd, raise_on_missing_sources=True)

        result = lookaside.sources(self.cwd, raise_on_missing_sources=False)
        self.assertEqual(result, None)

        open(source_path, 'w').close()
        lookaside.read_source_entries.return_value = [Munch({'filename': 'dist-git-1.11.tar.gz',
                                                             'hashtype': 'sha512',
                                                             'hash': '7aacb6'})]

        result = lookaside.sources('/repo_path', '/outdir')
        lookaside.download.assert_called_with(
            '/outdir/dist-git-1.11.tar.gz',
            '/repo_path',
            'dist-git-1.11.tar.gz',
            'sha512',
            '7aacb6')

        lookaside.download.reset_mock()
        result = lookaside.sources('/repo_path')
        lookaside.download.assert_called_with(
            os.path.join(self.cwd, 'dist-git-1.11.tar.gz'),
            '/repo_path',
            'dist-git-1.11.tar.gz',
            'sha512',
            '7aacb6')

        os.unlink(source_path)

    @patch('rpkglib.lookaside_cache.log')
    def test_upload(self, log):
        lookaside = LookasideCache(self.cwd, None, None)
        lookaside.make_source_entry = Mock()
        lookaside.write_source_entries = Mock()
        lookaside.upload_source = Mock()

        source_entries = [
            Munch({'filename': 'foo.tar.gz', 'hashtype': 'sha512', 'hash': '7aacb6'}),
            Munch({'filename': 'bar.tar.gz', 'hashtype': 'sha512', 'hash': '0cf919'}),
        ]

        lookaside.make_source_entry.side_effect = source_entries

        lookaside.upload('rpms/repo', ['/foo.tar.gz', '/bar.tar.gz'])
        self.assertEqual(lookaside.upload_source.mock_calls, [
            call('/foo.tar.gz', 'rpms/repo', 'foo.tar.gz', 'sha512', '7aacb6'),
            call('/bar.tar.gz', 'rpms/repo', 'bar.tar.gz', 'sha512', '0cf919')
        ])
