import subprocess

from mock import Mock, patch, call

from rpkglib import utils, errors
from . import base

class TestUtils(base.TestCase):

    def setUp(self):
        super(TestUtils, self).setUp()

    def tearDown(self):
        super(TestUtils, self).tearDown()

    def test_changelog_entry(self):
        commit_message = 'first line %changelog\nsecond_line'
        changelog_entry = utils.changelog_entry(commit_message)
        assert changelog_entry == '- first line %%changelog'

        commit_message = 'first line %changelog'*10
        changelog_entry = utils.changelog_entry(commit_message)
        split_lines = changelog_entry.split()
        for line in split_lines:
            assert len(line) <= 80

    @patch('subprocess.Popen')
    def test_run(self, popen):
        process = Mock()
        process.returncode = 0
        process.communicate.return_value = (None, b'err\n')
        popen.return_value = process

        utils.run(['echo', 'foo'])
        popen.assert_called_with(['echo', 'foo'], cwd=None,
                                 env=None, shell=False, stderr=None,
                                 stdout=subprocess.PIPE)
        process.communicate.assert_called()

        utils.run('echo foo', shell=True,
                  env={'var': 'val'}, cwd='/tmp',
                  capture=False, capture_stderr=True)

        popen.assert_called_with('echo foo', cwd='/tmp', env={'var': 'val'},
                                 shell=True, stderr=subprocess.PIPE,
                                 stdout=None)

        process.returncode = 1
        with self.assertRaises(errors.RpkgError) as e:
            utils.run(['echo', 'foo'])
            self.assertAlmostEquals(e.msg, 'err')

        result = utils.run(['echo', 'foo'], throw=False)
        self.assertEqual(result.rc, 1)

        process.returncode = 0
        process.communicate.return_value = (b'foo\n', None)

        result = utils.run(['echo', 'foo'])
        self.assertEqual(result.out, 'foo')
        self.assertEqual(result.err, '')
        self.assertEqual(result.rc, 0)

    @patch('os.unlink')
    @patch('subprocess.check_call')
    @patch('builtins.open')
    @patch('tempfile.NamedTemporaryFile')
    def test_edit(self, tmpfile_cls, open, check_call, unlink):
        tmpfile = Mock()
        tmpfile.name = 'tmpfile'
        tmpfile.read.return_value = 'hello world'
        tmpfile_cls.return_value = tmpfile
        open.return_value = tmpfile
        result = utils.edit('vim', 'hello world')
        tmpfile.assert_has_calls([call.write('hello world'), call.close(),
                                  call.read(), call.close()])
        check_call.assert_called_with(['vim', 'tmpfile'])
        open.assert_called_with('tmpfile', 'r')
        unlink.assert_called_with('tmpfile')
        self.assertEqual(result, 'hello world')

    def test_parse_config_rpmdefines(self):
        result = utils.parse_config_rpmdefines('"dist %nil"')
        self.assertEqual(result, ['dist %nil'])
        result = utils.parse_config_rpmdefines('\'dist %nil\'')
        self.assertEqual(result, ['dist %nil'])
        result = utils.parse_config_rpmdefines('\'dist %nil\' "release 1"')
        self.assertEqual(result, ['dist %nil', 'release 1'])

    def test_parse_config_path_list(self):
        result = utils.parse_config_path_list('"git.bash" /tmp/foo.bash \'/tmp/bar.bash\'')
        self.assertEqual(result, ['git.bash', '/tmp/foo.bash', '/tmp/bar.bash'])

    def test_cmd_repr(self):
        result = utils.cmd_repr(['git', 'clone', '--rebase', 'https://pagure.io/rpkg-util'])
        self.assertEqual(result, 'git clone --rebase https://pagure.io/rpkg-util')
        result = utils.cmd_repr(['rpmbuild', '--define', '_sourcedir foo', '-ba', '*.spec'])
        self.assertEqual(result, 'rpmbuild --define \'_sourcedir foo\' -ba \'*.spec\'')

    def test_has_scheme(self):
        result = utils.has_scheme('https://pagure.io/rpkg-util')
        self.assertEqual(result, True)
        result = utils.has_scheme('/tmp/foo')
        self.assertEqual(result, False)
        result = utils.has_scheme('tmp/foo')
        self.assertEqual(result, False)

    def test_macro_helper_cmd(self):
        result = utils.macro_helper_cmd('foo', '--bar')
        self.assertEqual(result, [utils.MACRO_HELPER_DIR+'foo', '--bar'])

    @patch('rpkglib.utils.run')
    def test_extract_srpm(self, run):
        result = utils.extract_srpm('/tmp/x.srpm', '/destdir')
        run.assert_called_with('rpm2cpio /tmp/x.srpm | cpio -idm --quiet', cwd='/destdir', shell=True)

    def test_is_upload_item(self):
        result = utils.is_upload_item('foo.txt')
        self.assertFalse(result)
        result = utils.is_upload_item('foo.tar.gz')
        self.assertTrue(result)
        result = utils.is_upload_item('foo.src.rpm')
        self.assertTrue(result)
        result = utils.is_upload_item('foo.cpp')
        self.assertFalse(result)

    def test_is_reserved_item(self):
        result = utils.is_reserved_item('.gitignore')
        self.assertTrue(result)
        result = utils.is_reserved_item('.git')
        self.assertTrue(result)
        result = utils.is_reserved_item('.gitattributes')
        self.assertTrue(result)
        result = utils.is_reserved_item('README.md')
        self.assertTrue(result)
        result = utils.is_reserved_item('README.txt')
        self.assertTrue(result)
        result = utils.is_reserved_item('rpkg.conf')
        self.assertTrue(result)
        result = utils.is_reserved_item('.rpkg')
        self.assertTrue(result)
        result = utils.is_reserved_item('.fedpkg', 'fedpkg')
        self.assertTrue(result)
        result = utils.is_reserved_item('fedpkg.conf', 'fedpkg')
        self.assertTrue(result)
        result = utils.is_reserved_item('fedpkg.conf', 'rpkg')
        self.assertFalse(result)
        result = utils.is_reserved_item('.fedpkg', 'rpkg')
        self.assertFalse(result)
        result = utils.is_reserved_item('foo.spec')
        self.assertFalse(result)
        result = utils.is_reserved_item('foo.tar.gz')
        self.assertFalse(result)
        result = utils.is_reserved_item('foo.crate')
        self.assertFalse(result)
